"""
PONGsim (Python Overlay Network Graphical Simulator)
Main module of the PONGsim framework

Version: 7.50

Nicklas Beijar, nbeijar@netlab.tkk.fi
Networking Laboratory, Helsinki University of Technology

######################################################################

PONGsim (Python Overlay Network Graphical Simulator)

Copyright (c) 2005-2007 Nicklas Beijar

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

######################################################################

To do:
- PyDoc compatible comments

"""

try:
    from Tkinter import *
except ImportError:
    pass

import random, string, thread, math, time, os, bisect, sys, traceback, os.path

WORLDSIZEX = 1000
WORLDSIZEY = 800

DEFAULTTRANSMISSIONDELAY = 0.020
DEFAULTINTERNALDELAY = 0.001

######################################################################
#####    S C E N A R I O
######################################################################

class Scenario:
    
    """Store the parameters of a simulated scenario."""

    def __init__(self, name, log=None):
        """Initialize a scenario."""
        self.data = {}
        self.name = name
        self.log = log
        
    def get(self, param, default=None):
        """Return the value of a given scenario parameter."""
        if self.data.has_key(param):
            return self.data[param]
        return default 

    def set(self, param, value):
        """Set the value of a given scenario parameter."""
        self.data[param] = value

    def set_typechk(self, param, value):
        """Set the value of a given scenario parameter by converting a string to a suitable type."""
        if value == "":
            value = None
        elif value[0]=='"' and value[-1]=='"':
            value = value[1:-1]
        elif value[0]=='{' and value[-1]=='}':
            value = value
        elif value == "True":
            value = True
        elif value == "False":
            value = False
        else:
            try:
                value = float(value)
            except ValueError:
                if self.log: self.log.warning("Non-recognized parameter type in '"+str(value)+"' - assuming string")
                value = value
        self.data[param] = value

    def getname(self):
        """Return the name of the scenario."""
        return self.name

    def getall(self):
        """Return the dictionary of all parameter values."""
        return self.data
    
    def getfilename(self):
        """Return the file name by parsing the file parameter and inserting values of specified parameters."""
        pattern = self.get("file", self.name)
        if pattern == None:
            if self.log: self.log.warning("No filename specified - using default")
        while string.find(pattern, "[") >= 0:
            start = string.find(pattern, "[")
            end = string.find(pattern, "]", start)
            if start < 0 or end < 0 or end<=start:
                if self.log: self.log.warning("Illegal name pattern: "+str(pattern))
                return self.name
            param = pattern[start+1:end]
            value = str(self.get(param, "_"+param+"_"))
            if value[-2:] == ".0": value = value[:-2]
            value = string.replace(value, " ", "")
            pattern = pattern[:start] + value + pattern[end+1:]
        return pattern
        
    def copy(self):
        """Return a new identical copy of the current scenario."""
        newscenario = Scenario(self.name, self.log)
        for param in self.data.keys():
            newscenario.set(param, self.data[param])
        return newscenario

    def findmultiline(self):
        """Return the first line with several values in the current scenario.
        
        Returns the (parameter, valuerange) of the first line with multiple values in this scenario or (None, None) if not found.
        Such a line is in the format "[value1 value2 value3]" defining several values for different simulator runs.
        """
        for param in self.data.keys():
            try:
                if self.data[param][0] == "{" and self.data[param][-1] == "}":
                    return param, self.data[param][1:-1]
            except TypeError:
                pass
        return None, None

    def __str__(self):
        """Return a string representation of this scenario."""
        t = ""
        keys = self.data.keys()
        keys.sort()
        for param in keys:
            t += str(param) + ";" + str(self.data[param]) + "\n"
        return t

######################################################################
#####    S C E N A R I O M A N A G E R
######################################################################

class ScenarioManager:
    
    """Load and store the scenarios, and manage the queue of scenarios to run."""
    
    def __init__(self, log):
        """Initiate the ScenarioManager."""
        self.scenarios = []
        self.activescenario = None
        self.activerun = 0
        self.log = log
        self.resetqueue()
        
    def loadall(self, path):
        """Load all scenarios in the given path."""
        for filename in os.listdir(path):
            if string.find(filename, ".scenario.txt") > -1: 
                self.load(path, filename)
            
    def load(self, path, filename):
        """Load a scenario in the given file in the given path."""
        scenario = string.replace(filename, ".scenario.txt", "")
        scenario = string.splitfields(scenario, "/")[-1]
        scenario = string.splitfields(scenario, "\\")[-1]
        self.log.status("Loading scenario "+scenario)
        fp = open(os.path.normcase(path + filename), "r")
        scen = Scenario(scenario, self.log)
        while 1:
            line = fp.readline()
            if not line: break
            line = string.replace(string.replace(line, "\r", ""), "\n", "")
            fields = string.splitfields(line, "=")
            if len(line) > 0 and line[0] != "#":
                if len(fields) == 2:
                    value = fields[1]
                    scen.set_typechk(fields[0], value)
                else:
                    self.log.warning("Skipping unknown line in "+filename+":\n"+line)
        fp.close()
        for multiscen in self.makemultiscenarios(scen):
            self.scenarios.append((scenario, multiscen))
        if self.activescenario == None: self.activescenario = scen

    def count(self):
        """Return the number of scenarios."""
        return len(self.scenarios)

    def makemultiscenarios(self, scenario):
        """Generate several scenarios from the given scenario, if there are lines specifying several values."""
        scenarios = []
        scenarios.append(scenario)
        changed = True
        while changed == True:
            changed = False
            for scenario in scenarios:
                param, value = scenario.findmultiline()
                if param != None:
                    changed = True
                    values = string.splitfields(value, " ")
                    scenario.set_typechk(param, values[0])
                    for value in values[1:]:
                        newscenario = scenario.copy()
                        newscenario.set_typechk(param, value)
                        scenarios.append(newscenario)
        return scenarios

    def enqueue(self, scenarioname, runnumber):
        """Put a given scenario on the queue and mark it with the given runnumber."""
        for (name, scen) in self.scenarios:
            if name == scenarioname:
                self.queue.append((runnumber, scen))

    def resetqueue(self):
        """Empty the queue."""
        self.queue = []

    def pop(self):
        """Return the first enqueued scenario and remove it from the queue."""
        if len(self.queue) == 0: return None
        (self.activerun, self.activescenario) = self.queue.pop(0)
        self.activescenario.set("runnumber", self.activerun)
        self.activescenario.set("rundate", time.strftime("%Y%m%d"))
        self.activescenario.set("runtime", time.strftime("%H%M"))
        self.activescenario.set("runoutput", self.activescenario.getfilename())
        return self.activescenario

    def getscenarios(self):
        """Return the names of all scenarios."""
        names = []
        for (name, scen) in self.scenarios:
            if name not in names: names.append(name)
        return names

    def getrunnumberXX(self):
        """Return the number of the currently active run."""
        return self.activerun

    # legacy
    def getXX(self, param, default=None):
        """Return the value of a given parameter of the currently active scenario."""
        if self.activescenario == None: return None
        return self.activescenario.get(param, default)

    def getactivescenario(self):
        """Return the currently active scenario."""
        return self.activescenario

######################################################################
#####    L O G
######################################################################

class Log:
    
    """Log for storing and displaying errors and other messages."""

    def __init__(self, scheduler, logfile = None, mode="a"):
        """Initialize the log and open the logfile (optional) for writing in the given mode (a=append, w=write)."""
        self.fp = None
        self.scheduler = scheduler
        if logfile:
            self.fp = open(os.path.normcase(logfile), mode)
    
    def simerror(self, msg):
        """Log an error message related to the simulation."""
        msg = "%s: ERROR: %s (sim.time=%1.3f)\n" % (time.strftime("%Y%m%d %H%M"), msg, self.scheduler.time())
        print msg,
        if self.fp: self.fp.write(msg)

    def simwarning(self, msg):
        """Log a warning message related to the simulation."""
        msg = "%s: WARNING: %s (sim.time=%1.3f)\n" % (time.strftime("%Y%m%d %H%M"), msg, self.scheduler.time())
        print msg,
        if self.fp: self.fp.write(msg)

    def simstatus(self, msg):
        """Log a status message related to the simulation."""
        msg = "%s: %s (sim.time=%1.3f)\n" % (time.strftime("%Y%m%d %H%M"), msg, self.scheduler.time())
        print msg,
        if self.fp: self.fp.write(msg)

    def status(self, msg):
        """Log a status message."""
        msg = "%s: %s\n" % (time.strftime("%Y%m%d %H%M"), msg)
        print msg,
        if self.fp: self.fp.write(msg)

    def error(self, msg):
        """Log an error message."""
        msg = "%s: ERROR: %s\n" % (time.strftime("%Y%m%d %H%M"), msg)
        print msg,
        if self.fp: self.fp.write(msg)

    def warning(self, msg):
        """Log a warning message."""
        msg = "%s: WARNING: %s\n" % (time.strftime("%Y%m%d %H%M"), msg)
        print msg,
        if self.fp: self.fp.write(msg)

    def close(self):
        """Close the log and its related file."""
        if self.fp: self.fp.close()
        
######################################################################
#####    C O U N T E R
######################################################################

class Counter:
    
    """Lightweitht statistics utility for counting events of different type."""

    def __init__(self, active = True, eventfile = None):
        self.counters = {}
        self.active = active
        self.fp = None
        if eventfile:
            self.fp = open(os.path.normcase(eventfile), "w")

    def activate(self):
        """Activate the counter."""
        self.active = True
        
    def deactivate(self):
        """Deactivate the counter."""
        self.active = False

    def add(self, name, value=1):
        """Increase the named counter with the given value (default 1)."""
        if not self.active: return
        try:
            self.counters[name] += value
        except KeyError:
            self.counters[name] = value
        if self.fp: self.fp.write("%1.4f;%s;%d\n" % (self.scheduler.time(), name, value))

    def add_override(self, name, value=1):
        """Increase the named counter with the given value (default 1) even if result collection has not started."""
        try:
            self.counters[name] += value
        except KeyError:
            self.counters[name] = value
        if self.fp: self.fp.write("%1.4f;%s;%d\n" % (self.scheduler.time(), name, value))

    def delete(self, name):
        """Delete the named counter."""
        try:
            del self.counters[name]
        except KeyError:
            pass

    def sum(self, name):
        """Return the value of the named counter, or None if it is not existing."""
        try:
            return self.counters[name]
        except KeyError:
            return None
        
    def counterlist(self):
        """Return the list of names of the counters."""
        return self.counters.keys()

    def close(self):
        """Close this Counter and its related file."""
        if self.fp: self.fp.close()

    def __str__(self):
        """Return a string representation of this Counter."""
        s = ""
        keys = self.counters.keys()
        keys.sort()
        for name in keys:
            s += str(name) + ";" + str(self.counters[name]) + "\n"
        return s

######################################################################
#####    S T A T I S T I C S C O U N T E R
######################################################################

class StatisticsCounter:
    
    """More heavy statistics utility for counting events of different type with analysis of the values.
    
    Calculates sum, max, min, average and number of entries. Compatible with Counter.
    """

    def __init__(self, active = True, eventfile = None):
        self.counters = {}
        self.active = active
        self.fp = None
        if eventfile:
            self.fp = open(os.path.normcase(eventfile), "w")

    def activate(self):
        """Activate the counter."""
        self.active = True
        
    def deactivate(self):
        """Deactivate the counter."""
        self.active = False

    def add(self, name, value=1):
        """Increase the named counter with the given value (default 1)."""
        if not self.active: return
        try:
            self.counters[name] = (self.counters[name][0]+1, self.counters[name][1]+value, min(self.counters[name][2], value), max(self.counters[name][3], value))
        except KeyError:
            self.counters[name] = (1, value, value, value)
        if self.fp: self.fp.write("%1.4f;%s;%d\n" % (self.scheduler.time(), name, value))

    def add_override(self, name, value=1):
        """Increase the named counter with the given value (default 1) even if result collection has not started."""
        try:
            self.counters[name] += value
        except KeyError:
            self.counters[name] = value
        if self.fp: self.fp.write("%1.4f;%s;%d\n" % (self.scheduler.time(), name, value))

    def delete(self, name):
        """Delete the named counter."""
        try:
            del self.counters[name]
        except KeyError:
            pass

    def sum(self, name):
        """Return the value of the named counter, or None if it is not existing."""
        try:
            return self.counters[name][1]
        except KeyError:
            return None

    def average(self, name):
        """Return the value of the named counter, or None if it is not existing."""
        try:
            return float(self.counters[name][1])/float(self.counters[name][0])
        except KeyError:
            return None

    def entrycount(self, name):
        """Return the value of the named counter, or None if it is not existing."""
        try:
            return self.counters[name][0]
        except KeyError:
            return None

    def max(self, name):
        """Return the value of the named counter, or None if it is not existing."""
        try:
            return self.counters[name][3]
        except KeyError:
            return None

    def min(self, name):
        """Return the value of the named counter, or None if it is not existing."""
        try:
            return self.counters[name][2]
        except KeyError:
            return None

    def counterlist(self):
        """Return the list of names of the counters."""
        return self.counters.keys()

    def close(self):
        """Close this Counter and its related file."""
        if self.fp: self.fp.close()

    def __str__(self):
        """Return a string representation of this Counter."""
        s = ""
        keys = self.counters.keys()
        keys.sort()
        for name in keys:
            s += str(name) + ";" + str(self.counters[name][0]) + ";" + str(self.counters[name][1]) + ";" + str(self.counters[name][2]) + ";" + str(self.counters[name][3]) + ";" + str(float(self.counters[name][1]) / float(self.counters[name][0])) + "\n"
        return s

######################################################################
#####    S N A P S H O T
######################################################################

class Snapshot:
    
    """Statistics utility for making periodical snapshots of the current simulated state."""
    
    def __init__(self, scheduler, interval, callback, filename):
        """Initialize the Snapshot.
        
        Arguments:
        scheduler - Sheduler for scheduling the snapshot event.
        interval - Interval in simulated sime for performing the snapshot.
        callback - Callback function that is called for obtaining the snapshot value.
        filename - File to which the snapshot is written.
        
        The callback function is a function without arguments that returns any type of value.
        """
        self.scheduler = scheduler
        self.filename = filename
        self.callback = callback
        self.fp = None
        if self.filename:
            self.fp = open(os.path.normcase(self.filename), "w")
        if not self.fp:
            self.events = []
        scheduler.every(interval, self.event)

    def event(self, dummy):
        """Handles a snapshot event (internal function)."""
        event = str(self.callback())
        if self.fp: 
            self.fp.write("%1.4f;%s\n" % (self.scheduler.time(), event))
        else:
            self.events.append((self.scheduler.time(), event, None))
    
    def close(self):
        """Closes the Snapshot and its related file."""
        self.fp.close()

######################################################################
#####    E V E N T Q U E U E
######################################################################

class EventQueue:
    
    """Queue for simulated events."""
    
    def __init__(self):
        """Initialize the EventQueue"""
        self.queue = []
    
    def enqueue(self, eventtime, function, parameters, repeat):
        """Enqueue a given function at the location determined by the given time."""
        bisect.insort(self.queue, (eventtime, function, parameters, repeat))

    def dequeue(self):
        """Return the following event from the queue and remove it from the queue."""
        return self.queue.pop(0)
        
    def length(self):
        """Return the length of the queue."""
        return len(self.queue)

######################################################################
#####    S C H E D U L E R
######################################################################

class Scheduler:
    
    """Scheduler for storing simulated events and for managing simulated time."""
    
    def __init__(self, log=None, until=None):
        """Initialize the scheduler."""
        self.until = until
        self.log = log
        self.reset()
        
    def setlog(self, log):
        """Sets the Log to which warnings are written."""
        self.log = log
        
    def reset(self):
        """Reset the scheduler for a new simulation."""
        self.eventqueue = EventQueue()
        self.simtime = 0.0
        
    def run(self):
        """Run all scheduled events."""
        while 1:
            if step(): break
        
    def step(self):
        """Run a single scheduled event.
        
        Returns True if the last event was run, otherwise returns False. 
        """
        if (self.until != None and self.simtime > self.until): return True
        try:
            (eventtime, function, parameters, repeat) = self.eventqueue.dequeue()
        except IndexError:
            return True
        if eventtime < self.simtime:
            if self.log: self.log.simwarning("Skipping event scheduled in the past (eventtime="+str(eventtime)+") ")
        else:
            self.simtime = eventtime
            if repeat:
                self.eventqueue.enqueue(self.simtime + repeat, function, parameters, repeat)
            #print function, parameters
            function(parameters)
        return False
        
    def at(self, eventtime, function, parameters=None):
        """Schedule a given event at a given time."""
        self.eventqueue.enqueue(eventtime, function, parameters, None)
        
    def after(self, eventtime, function, parameters=None):
        """Schedule a given event after the given time."""
        if eventtime < 0: 
            raise Application
            if self.log: self.log.simwarning("Skipping scheduling event after negative time (after="+str(eventtime)+")")
            return
        if eventtime == 0: 
            if self.log: self.log.simwarning("Event scheduled immediately - accepted but unrealistic")
        self.eventqueue.enqueue(self.simtime + eventtime, function, parameters, None)
        
    def every(self, eventtime, function, parameters=None):
        """Schedule a given event to be run repeated at a given interval."""
        self.eventqueue.enqueue(self.simtime + eventtime, function, parameters, eventtime)
        
    def time(self):
        """Return the current simulated time."""
        return self.simtime
        
    def queuelength(self):
        """Return the current number of pending events."""
        return self.eventqueue.length()

######################################################################
#####    A P P L I C A T I O N
######################################################################

class Application:
    
    """A framework of a simulated application.
    
    This class provides the framework for overlay protocols and applications. These should
    be implemented as subclasses of the Application class. 
    """
    
    def __init__(self, node):
        """Initiate the application.
        
        The original function MUST be called from an overriding function. A better approach
        is to override the init() function. 
        """
        self.node = node
        self.node.setapplication(self)
        self.simulation = self.node.world.simulation
        self.init()
        
    def init(self):
        """Initiate the application.
        
        Implement your application by overriding this function.
        """
        pass
        
    def send(self, message, destination=None):
        """Send the given message to the given destination."""
        self.node.send(message, destination)

    def receive(self, message):
        """Handle a received message.
        
        Implement your application by overriding this function.
        """
        pass

    def getdisplayinfo(self):
        """Return the text to be written beside the node.
        
        Implement your application by overriding this function. Call the overrided function to include the 
        default info. 
        """
        return self.node.getdisplayinfo()

    def getdetailedinfo(self):
        """Return the the node's description that is written in a separate window when the node is examined.
        
        Implement your application by overriding this function. Call the overrided function to include the 
        default info. 
        """
        return self.node.getdetailedinfo()

    def draw(self, canvas, x, y):
        """Draw the graphical representation of this application. 
        
        Implement your application by overriding this function.
        """
        pass

######################################################################
#####    N O D E
######################################################################

class Node:
    
    """Class representing a simulated node.
    
    An application can be attached to the node. A node must belong to a World.
    """
    
    nextaddress = 1
    
    ##### INITIALIZATION

    def __init__(self, world, address=None, noderadius=5, color="#777", outline="#000", x=None, y=None):
        """Initialize this node.
        
        Arguments:
        world - The World to which the node belongs.
        address - The address of the node. If None is specified, an automatically generated address is assigned. Default None. 
        noderadius - The radius of the drawn node. Default 5.
        color - The fill color of the drawn node. Default "#777".
        outline - The outline color of the drawn node. Default "#000".
        x - The x-coordinate of the drawn node. If None is specified, the coordinate is random. Default None.
        y - The yx-coordinate of the drawn node. If None is specified, the coordinate is random. Default None.
        """
        if x:
            self.x = x
        else:
            self.x = random.randint(0, world.xsize)
        if y:
            self.y = y
        else:
            self.y = random.randint(0, world.ysize)
        if address == None:
            self.address = Node.nextaddress
            Node.nextaddress += 1
        else:
            self.address = address
        self.noderadius = noderadius
        self.fill = color
        self.outline = outline
        self.application = None
        self.world = world
        self.world.addnode(self)
        self.counter = Counter(True)

    def setapplication(self, application):
        """Attach the given Application to this node.
        
        NOTE: This is called automatically by the Application as part of its initialization."""
        self.application = application

    ##### NODE-WORLD RELATIONSHIP

    def remove(self):
        """Remove this node."""
        assert self.world != None
        del self.world.nodes[self.world.nodes.index(self)]
        assert self not in self.world.nodes
        self.world = None
        
    def inworld(self):
        """Returns True if this node is in the world, i.e. if the node exists."""
        return self.world != None

    ##### INFORMATION RETRIEVAL

    def __str__(self):
        """Return the address as the string representation of this node."""
        return str(self.address)

    def __repr__(self):
        """Return the identifier of the node as the representation of this node."""
        return "Node"+str(self.address)

    def getdetailedinfo(self):
        """Return the the node's description that is written in a separate window when the node is examined.
        
        Implement your application by overriding this function. Call the overrided function to include the 
        default info. 
        """
        t = "Address: "+str(self.address)+"\nCounters:\n" + str(self.counter)
        return t

    def getdisplayinfo(self):
        """Return the text to be written beside the node.
        
        Implement your application by overriding this function. Call the overrided function to include the 
        default info. 
        """
        return str(self.address)

    def getaddress(self):
        """Return the address of this node."""
        return self.address

    ##### NODE-SPECIFIC COUNTERS

    def count(self, name, value=1):
        """Increase the named node-specific counter with the given value (default 1).
        
        For compatibility.
        """
        self.counter.add(name, value)

    def count_override(self, name, value=1):
        """Increase the named node-specific counter with the given value (default 1) even if result collection has not started.
        
        For compatibility.
        """
        self.counter.add_override(name, value)

    ##### DISPLAY

    def getx(self):
        """Return the x-coordinate of this node."""
        return self.x

    def gety(self):
        """Return the y-coordinate of this node."""
        return self.y

    def getnoderadius(self):
        """Return the radius of this node."""
        return self.noderadius

    def draw(self, canvas):
        """Draw this node on the given canvas."""
        assert self.world
        canvas.create_oval(self.x-self.noderadius, self.y-self.noderadius, self.x+self.noderadius, self.y+self.noderadius, fill=self.fill, outline=self.outline)
        if self.application: 
            s = self.application.getdisplayinfo()
        else:
            s = self.getdisplayinfo()
        canvas.create_text(self.x+2*self.noderadius, self.y, text=s, anchor=W)
        if self.application: self.application.draw(canvas, self.x, self.y)

    ##### MESSAGES
    
    def receive(self, message):
        """Handles reception of the given message.
        
        Forwards the received message to the attached Application.
        """
        if not self.world: 
            self.count("lost-msg-receiver-not-in-world")
            return
        self.count("received-msg")
        if self.application: self.application.receive(message)
            
    def send(self, message, destination, delay=DEFAULTTRANSMISSIONDELAY, internaldelay=DEFAULTINTERNALDELAY):
        """Sends the given message to the given destination with the given delay (default 20ms).
        
        If the message is sent to the node itself, the internal delay (default 1ms) is applied instead of the transmission delay.
        """
        if not self.world: 
            self.count("error-activity-by-nonexistent-node")
            return
        self.count("sent-msg")
        if destination == self:
            self.world.simulation.scheduler.after(internaldelay, destination.receive, message)
        else:
            self.world.simulation.scheduler.after(delay, destination.receive, message)

######################################################################
#####    W O R L D 
######################################################################

class World:
    
    """Class for storing the nodes in a simulation.
    
    Every Node belongs to a World. The simulator simulates a World. 
    The concept of World is a bit outdated in the current version of the simulator, but is included
    for compatibility. 
    """
    
    def __init__(self, simulation, xsize=WORLDSIZEX, ysize=WORLDSIZEY):
        """Initiate the World. 
        
        The xsize and ysize arguments specify the graphical size of the world. 
        """
        self.simulation = simulation
        self.xsize = xsize
        self.ysize = ysize
        self.nodes = []
        
    def addnode(self, node):
        """Add a Node to the World.
        
        NOTE: This function is called by the Node automatically.
        """
        if node not in self.nodes: self.nodes.append(node)
        
    def clear(self):
        """Remove all nodes from the world."""
        for node in self.nodes:
            node.world = None
        self.nodes = []
        
    def randomnode(self):
        """Returns a random Node."""
        return self.nodes[random.randint(0, len(self.nodes)-1)]
    
    def randomcoordinate(self):
        """Return random (x, y) coordinates within the graphical size of the world."""
        return (random.randint(0, self.xsize), random.randint(0, self.ysize))

    def getnodes(self):
        """Return a list of all nodes in the world."""
        return self.nodes[:] # make a new list

    def draw(self, canvas):
        """Draw the world on the given canvas."""
        canvas.delete(ALL)
        for node in self.nodes:
            node.draw(canvas)

    def nodecount(self):
        """Return the number of nodes in the world."""
        return len(self.nodes)

    def closest(self, x, y):
        """Return the Node that is closest to the given x- and y-coordinates."""
        closest = None
        closestdist = 2*(self.xsize+self.ysize)
        for node in self.nodes:
            d = math.sqrt((x - node.x)*(x - node.x) + (y - node.y)*(y - node.y))
            if d < closestdist: 
                closest = node
                closestdist = d
        return closest

######################################################################
#####    S I M U L A T I O N
######################################################################

class Simulation:

    def __init__(self):
        """Initialize the simulation
        
        This function is called when the simulation is being initialized. 
        Override this function to implement your simulator. 
        """
        self.scenario = None    # set by Simulator
        self.scheduler = None    # set by Simulator
        self.log = None    # set by Simulator
        self.finish = False # set by the user to stop the simulation, read by Simulator, reset by simulator
        self.world = None # set by makeworld()

    def makeworld(self):
        """Return a World for the simulation.
        
        Called before init(). This function CAN be overrided to define other types of Worlds
        """
        self.world = World(self, WORLDSIZEX, WORLDSIZEY)
        return self.world

    def eachevent(self):
        """Perform actions after each event.
        
        A common use of this function is to define when to stop (e.g. based on time).
        Override this function to implement your simulator. 
        """
        pass

    def init(self):
        """Initialize a simulation run.
        
        This function is called when the simulation run is being initialized. 
        Override this function to implement your simulator. 
        """
        pass

    def close(self):
        """Close the simulation.
        
        This function is called when the simulation has been finished or stopped. It is a suitable place
        to collect results from the simulation. 
        Override this function to implement your simulator. 
        """
        pass

######################################################################
#####    S T A T U S F I E L D
######################################################################

class StatusField:
    
    """User-extendable display of various status information about the simulation."""
    
    
    def __init__(self, frame):
        """Initialise the StatusField and attach it to the given frame."""
        self.fields = []
        self.frame = frame
        
    def add(self, name, callback):
        """Add a field with a given name and a given callback function that returns the value to display."""
        namelabel = Label(self.frame, text=name, relief=RIDGE, width=25)
        namelabel.pack(side=TOP)
        label= Label(self.frame, text="-")
        label.pack(side=TOP)
        self.fields.append((label, callback))
        
    def draw(self):
        """Draw the StatusField."""
        for (label, callback) in self.fields:
            label.config(text=str(callback()))

######################################################################
#####    S I M U L A T O R
######################################################################

class Simulator:
    
    ##### INITIALIZATION
    
    STATE_IDLE = 1
    STATE_RUNNING = 2
    STATE_PAUSE = 3
    
    def __init__(self, simulation, logfilename, scenariodir = ".\\"):
        """Initialize the Simulation.
        
        Arguments:
        scheduler - An instance of the Scheduler class.
        master - The Tk root window.
        scenariomanager - An instance of the ScenarioManager class.
        """
        self.sim_mutex = thread.allocate_lock()
        self.scheduler = Scheduler()
        self.simulation = simulation
        self.log = Log(self.scheduler, logfilename)
        self.log.status("Start PONGsim")
        self.scheduler.setlog(self.log)
        self.simulation.log = self.log
        self.scenariomanager = ScenarioManager(self.log)
        self.infowindow = None
        self.exceptionwindow = None
        self.selectednode = None
        self.world = None
        self.state = self.STATE_IDLE
        self.eventspersecond = 0
        self.lastinfotext = ""
        self.stoprequest = False
        self.pauserequest = False
        self.counters = {}
        self.scenariodir = scenariodir
        self.interactive = True
        
        # parameters selected through UI:
        self.redrawinterval = 0.2
        self.delay = 0
        
    def run(self):
        """Run the simulator."""
        if len(sys.argv) > 2:
            self.runbatch(int(sys.argv[1]), sys.argv[2:])
        else:
            self.runinteractive()
    
    def runbatch_old(self, runtimes, scenarios):
        """Run the given scenarios in a batch mode without GUI."""
        self.interactive = False
        available = self.scenariomanager.getscenarios()
        errors = False
        for runnumber in range(runtimes):
            for scenario in scenarios:
                if scenario in available:
                    self.scenariomanager.enqueue(scenario, runnumber+1)
                else:
                    if runnumber == 0: self.log.error("Unknown scenario:"+scenario)
                    errors = True
        if errors:
            self.log.status("Available scenarios: "+string.joinfields(available, ", "))
        if not errors:
            self.runsim(1)
            self.close()

    def runbatch(self, runtimes, scenarios):
        """Run the given scenarios in a batch mode without GUI."""
        self.interactive = False
        for scenario in scenarios:
            self.scenariomanager.load("", scenario)
        available = self.scenariomanager.getscenarios()
        for runnumber in range(runtimes):
            for scenario in available:
                self.scenariomanager.enqueue(scenario, runnumber+1)
        self.log.status("Enqueued scenarios: "+string.joinfields(available, ", "))
        self.runsim(1)
        self.close()

    def runinteractive(self):
        """Run in an interactive mode with a GUI."""
        self.interactive = True
        self.scenariomanager.loadall(self.scenariodir)
        self.initscreen()
        self.master.mainloop()

    ##### CALLBACKS FOR STATUS
    
    def getscenario(self):
        """Return the name (including run number) of the running scenario."""
        if self.state == self.STATE_RUNNING or self.state == self.STATE_PAUSE:
            scenario = self.scenariomanager.getactivescenario().getname() + "  ("+str(self.scenariomanager.getactivescenario().get("runnumber")) + ")"
        else:
            scenario = "(None)"
        return scenario

    def getoutfile(self):
        """Return the name of the current output file."""
        if self.state == self.STATE_RUNNING or self.state == self.STATE_PAUSE:
            filename = string.replace(self.scenariomanager.getactivescenario().getfilename(), ".output.txt", "")
        else:
            filename = "(None)"
        return filename

    def getsimulatedtime(self):
        """Return the current simulated time."""
        return "%1.2fs (%1dh %1dm %1ds)" % (self.scheduler.time(), int(self.scheduler.time() / 3600), int((self.scheduler.time() % 3600) /60), self.scheduler.time() % 60)

    def geteventspersecond(self):
        """Return the current number of simulated events per second."""
        return int(self.eventspersecond)

    ##### DRAWING
    
    def initscreen(self):
        """Create the contents of the window and start periodical redrawing."""
        
        self.master = Tk()
        self.master.title("PONGsim")

        self.frame = Frame(self.master)
        self.frame.pack()
        rightframe = Frame(self.frame)
        rightframe.pack(side=RIGHT)
        buttonframe = Frame(rightframe)
        buttonframe.pack(side=TOP)
        
        self.statusfield = StatusField(buttonframe)
        self.statusfield.add("Running scenario:", self.getscenario)
        self.statusfield.add("Output file:", self.getoutfile)
        self.statusfield.add("Simulated time:", self.getsimulatedtime)
        self.statusfield.add("Events per second:", self.geteventspersecond)
        self.statusfield.add("Event queue length:", self.scheduler.queuelength)
        
        emptylabel = Label(buttonframe, text=" ")
        emptylabel.pack(side=TOP)
        
        scenarioframe = Frame(buttonframe)
        scenarioframe.pack(side=TOP)
        self.label1 = Label(scenarioframe, text="Available scenarios:", width=25)
        self.label1.pack(side=TOP)
        self.yScroll  =  Scrollbar (scenarioframe, orient=VERTICAL)
        self.yScroll.pack(side=RIGHT, fill=Y)
        self.listbox = Listbox (scenarioframe, yscrollcommand=self.yScroll.set, selectmode=MULTIPLE)
        self.listbox.pack(side=RIGHT, fill=X)
        scenarios = self.scenariomanager.getscenarios()
        scenarios.sort()
        for scenario in scenarios:
            self.listbox.insert(END, scenario)
        self.yScroll["command"]  =  self.listbox.yview
        self.listbox.bind("<ButtonRelease-1>", self.button_selectscenario)

        self.label5 = Label(buttonframe, text="Run times:", width=25)
        self.label5.pack(side=TOP)
        self.timesentry = Entry(buttonframe, width=5)
        self.timesentry.insert("1", 1)
        self.timesentry.pack(side=TOP)
        self.redrawintervalslider = Scale(buttonframe, label="Redraw interval (s)", resolution=0.10, orient=HORIZONTAL, sliderlength=20, length=140, from_=0.10, to=2, command=self.setredrawinterval)
        self.redrawintervalslider.pack(side=TOP)
        self.redrawintervalslider.set(self.redrawinterval)
        self.delayslider = Scale(buttonframe, label="Step delay (s)", resolution=0.0005, orient=HORIZONTAL, sliderlength=20, length=140, from_=0.0, to=0.00500, command=self.setdelay)
        self.delayslider.pack(side=TOP)
        self.delayslider.set(self.delay)

        emptylabel = Label(buttonframe, text=" ")
        emptylabel.pack(side=TOP)
        
        self.button1 = Button(buttonframe, text="Run", width=25, command=self.button_start)
        self.button1.pack(side=TOP)
        self.button2 = Button(buttonframe, text="Stop", width=25, command=self.button_stop)
        self.button2.pack(side=TOP)
        self.button3 = Button(buttonframe, text="Step", width=25, command=self.button_step)
        self.button3.pack(side=TOP)
        self.button4 = Button(buttonframe, text="Quit", fg="red", width=25, command=self.exit)
        self.button4.pack(side=TOP)

        self.canvas = Canvas(self.frame, height=800, width=1000, bg="white")
        self.canvas.bind("<Button-1>", self.clickleft)
        self.canvas.bind("<Button-2>", self.clickright)
        self.canvas.pack(side=TOP)

        self.master.after(1, self.redraw)
    
    def redraw(self):
        """Redraw the screen and reshedule this function."""
        self.sim_mutex.acquire()
        self.draw()
        self.sim_mutex.release()
        self.master.after(int(self.redrawinterval*1000), self.redraw) # give time to Tk to update window

    def draw(self, dummy=None):
        """Draw the World, the selected node, the status and the detail window."""
        if not self.world: return
        self.world.draw(self.canvas)
        if self.selectednode and self.selectednode.inworld(): self.canvas.create_oval(self.selectednode.getx()-self.selectednode.getnoderadius()-2, self.selectednode.gety()-self.selectednode.getnoderadius()-2, self.selectednode.getx()+self.selectednode.getnoderadius()+2, self.selectednode.gety()+self.selectednode.getnoderadius()+2, outline="#F00")

        self.statusfield.draw()

        if self.selectednode:
            if self.selectednode.application:
                if self.infowindow:
                    t = self.selectednode.application.getdetailedinfo()
                    if t != self.lastinfotext:
                        try:
                            # self.infolabel.config(text=t)
                            self.infotext.delete("1.0", END)
                            self.infotext.insert(END, t)
                        except TclError:
                            pass
                    self.lastinfotext = t

    ##### SIMULATOR

    def runsim(self, dummy):
        """Run the simulator.
        
        Should be run in a separate thread. Controlled externally through the following variables:
        self.stoprequest - Set True to stop the current simulation.
        self.pauserequest - Set True to pause the current simulation and False to continue after pausing.
        self.delay - Specify the delay to wait after each simulated event.
        """
        try:
            self.state = self.STATE_RUNNING
            activescenario = self.scenariomanager.pop()
            lastpasstime = None
            while activescenario != None and not self.stoprequest: 
                self.log.status("Initialize scenario: "+activescenario.getname()+" ("+str(activescenario.get("runnumber")) + ")")
                
                try:
                    self.sim_mutex.acquire()
                    self.reset()
                    self.sim_mutex.release()
                except:
                    self.exception("initialization")
                stop = False
                events = 0
                self.log.status("Run scenario: "+activescenario.getname()+" ("+str(activescenario.get("runnumber")) + ")")
                while not stop:
                    while self.pauserequest:
                        time.sleep(1)
                        lastpasstime = None
                    self.sim_mutex.acquire()
                    for ev in range(100):
                        if self.delay > 0: time.sleep(self.delay)
                        try:
                            outofevents = self.scheduler.step()
                            self.simulation.eachevent()
                        except:
                            self.exception("event handling in user code")
                            outofevents = False
                        if outofevents: 
                            stop = True
                            self.log.status("Finish scenario: "+activescenario.getname() + " ("+str(activescenario.get("runnumber")) + ") - no events left")
                            break
                        if self.simulation.finish:
                            stop = True
                            self.log.status("Finish scenario: "+activescenario.getname() + " ("+str(activescenario.get("runnumber")) + ") - finish")
                            break
                        if self.stoprequest or self.pauserequest:
                            if self.stoprequest:
                                stop = True
                                self.log.status("Abort scenario: "+activescenario.getname() + " ("+str(activescenario.get("runnumber")) + ") - stop requested")
                            break
                        events += 1
                    if lastpasstime and (time.time()-lastpasstime >0):
                        eventspersec = events / (time.time()-lastpasstime)
                        self.eventspersecond = self.eventspersecond*0.70 + 0.30*(eventspersec)
                    lastpasstime = time.time()
                    self.sim_mutex.release()
                self.simulation.close()
                activescenario = self.scenariomanager.pop()
            if self.stoprequest:
                self.log.status("Abort simulation: remaining scenarios skipped")
            else:
                self.log.status("Finish simulation: all scenarios finished")
            self.state = self.STATE_IDLE
            if self.interactive: self.button1.config(text="Run")
            self.stoprequest = False
        except:
            self.exception("simulator framework during simulator run")

    def enqueuescenarios(self):
        """Enqueue the scenarios selected in the listbox."""
        try:
            runtimes = int(self.timesentry.get())
        except ValueError:
            runtimes = 1
        self.scenariomanager.resetqueue()
        for i in range(runtimes):
            items = self.listbox.curselection()
            try:
                items = map(string.atoi, items)
            except ValueError: 
                pass
            for selection in items:
                self.scenariomanager.enqueue(self.listbox.get(selection), i+1)
        self.log.status("Start simulation: "+str(self.scenariomanager.count())+" scenarios, each "+str(runtimes)+ " times")

    def exception(self, where):
        """Show the stack dump on a program exception.
        
        In interactive mode, the dump is shown in a window, otherwise as text. 
        Argument where specifies a text describing where the exception happened.
        The simulator is paused.
        """
        text = string.joinfields(traceback.format_exception(sys.exc_type, sys.exc_value, sys.exc_traceback), "\n")
        self.log.error("Program error in "+where+"\n"+text+"\n")
        if self.interactive:
            text = "ERROR in " + where + "\n" + "-"*80+ "\n" + text + "-"*80+ "\n"
            text += "Simulation paused - press Continue to ignore and continue"
            self.closeexceptionwin()
            self.exceptionwindow = Toplevel(height=400, width=500)
            self.exceptionwindow.title("PONGsim - ERROR")
            frame = Frame(self.exceptionwindow, relief=RIDGE)
            frame.pack(side=LEFT)
            #label = Label(frame, text=text, height=40, width=80, justify=LEFT)
            #abel.pack(side=LEFT)    
            
            infotext = Text(frame, height=60, width=60, wrap=WORD)
            infotext.pack(side=LEFT)
            infotext.insert(END, text)
            scroll = Scrollbar(frame, orient=VERTICAL)
            scroll.pack(side=RIGHT, fill=Y)
            scroll.config(command=infotext.yview)
            infotext["yscrollcommand"] = scroll.set
            
            self.state = self.STATE_PAUSE
            self.button1.config(text="Continue")
            self.pauserequest = True
        else:
            self.stoprequest = True
            
    def closeexceptionwin(self):
        """Close the error window."""
        if self.exceptionwindow:
            try:
                self.exceptionwindow.withdraw()
            except TclError:
                pass
            self.exceptionwindow = None
            
    def reset(self):
        """Reset the simulator and initialize a new scenario."""
        self.scheduler.reset()
        self.eq = EventQueue()
        self.eventspersecond = 0
        self.selectednode = None
        self.simulation.scenario = self.scenariomanager.getactivescenario()
        self.simulation.scheduler = self.scheduler
        self.simulation.finish = False
        self.world = self.simulation.makeworld()
        self.simulation.init()

    def close(self):
        """Closes the simulator."""
        self.log.close()

   ##### USER INTERFACE EVENTS

    def exit(self):
        """Handle pressing the exit button."""
        self.closeinfowin()
        self.closeexceptionwin()
        self.frame.quit()
        self.close()

    def button_start(self):
        """Handle pressing of the start/pause/continue button."""
        self.closeexceptionwin()
        self.closeinfowin()
        if self.state == self.STATE_IDLE and not self.stoprequest:          # Start
            self.state = self.STATE_RUNNING
            self.button1.config(text="Pause")
            self.pauserequest = False
            self.stoprequest = False
            self.enqueuescenarios()
            thread.start_new_thread(self.runsim, (self,))
        elif self.state == self.STATE_PAUSE and not self.stoprequest:    # Resume
            self.state = self.STATE_RUNNING
            self.button1.config(text="Pause")
            self.pauserequest = False
        elif self.state == self.STATE_RUNNING and not self.stoprequest:  # Pause
            self.state = self.STATE_PAUSE
            self.button1.config(text="Continue")
            self.pauserequest = True

    def button_stop(self):
        """Handle pressing of the stop button."""
        self.closeexceptionwin()
        self.closeinfowin()
        if self.state == self.STATE_RUNNING or self.state == self.STATE_PAUSE:
            self.stoprequest = True

    def button_step(self):
        """Handle pressing of the step button."""
        self.closeexceptionwin()
        self.closeinfowin()
        if self.state == self.STATE_PAUSE:
            self.log.status("Step start")
            self.scheduler.step()
            self.simulation.eachevent()
            self.log.status("Step end")
               
    def setredrawinterval(self, event):
        """Handle moving the redraw interval slider."""
        self.redrawinterval = self.redrawintervalslider.get()

    def setdelay(self, event):
        """Handle moving the delay slider."""
        self.delay = self.delayslider.get()

    def button_selectscenario(self, event):
        """Handle selecting scenario in the scenario listbox."""
        pass
        
    def clickright(self, event):
        """Handle clicking the right mouse button in the world view."""
        node = self.world.closest(event.x, event.y)
        if node: 
            self.selected = node
            self.draw() # update info since automatic redrawing is disable
    
    def clickleft(self, event):
        """Handle clicking the left mouse button in the world view."""
        self.selectednode = self.world.closest(event.x, event.y)
        if self.infowindow == None: 
            self.infowindow = Toplevel(height=60, width=40)
            self.infoframe = Frame(self.infowindow, relief=RIDGE)
            self.infoframe.pack(side=LEFT)
            #self.infolabel = Label(self.infoframe, text="Waiting for info...\n\n", height=60, width=60, justify=LEFT)
            #self.infolabel.pack(side=LEFT)
            self.infotext = Text(self.infoframe, height=60, width=60, wrap=WORD)
            self.infotext.pack(side=LEFT)
            self.infotext.insert(END, self.selectednode.application.getdetailedinfo())
            scroll = Scrollbar(self.infoframe, orient=VERTICAL)
            scroll.pack(side=RIGHT, fill=Y)
            scroll.config(command=self.infotext.yview)
            #self.infowindow.bind("<Destroy>", self.closeinfowin)
            self.infotext["yscrollcommand"]  =  scroll.set
    
    def closeinfowin(self, event=None):    
        """Handle pressing the close window button of the info window."""
        if self.infowindow:
            try:
                self.infowindow.withdraw()
            except TclError:
                pass
            self.infowindow = None

if __name__ == '__main__':
    print "This is a framework for simulations, and is not runnable as such."
    print "To use it, create subclasses of the framework classes."
    