"""
Simulator for distribution algorithms for fully distributed user groups 
Based on the PONGsim framework

Version: 1.50

Nicklas Beijar, nbeijar@netlab.tkk.fi
Networking Laboratory, Helsinki University of Technology

######################################################################

Simulator for distribution algorithms for fully distributed user groups

Copyright (c) 2005-2007 Nicklas Beijar

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

######################################################################

To do:
- Everything

"""

import random, math, time, os.path
import pongsim, pongsim_topologygenerator, pongsim_resources, pongsim_ri_app, pongsim_floodsearch_app, pongsim_floodindex_app

####################################################################
###   G r o u p S i m u l a t i o n
####################################################################

class GroupSimulation(pongsim.Simulation):

    def __init__(self, outputdir):
        """Initialize the simulation
        
        This function is called when the simulation is being initialized. 
        """
        pongsim.Simulation.__init__(self)
        self.outputdir = outputdir
        
    def init(self):
        """Initiate the user-part of the simulation."""
        global sm

        reportname = self.outputdir + self.scenario.get("runoutput")
        self.noreport = False
        try:
            fp = open(os.path.normcase(reportname), "r")
            self.noreport = True        # if file exists, don't simulate again
            self.finish = True
            self.log.simstatus("Already existing -> Skipping")
        except:
            pass

        self.simcounter = pongsim.StatisticsCounter()
        sm = self.scenario
        random.seed(sm.get("randomseed", 0) + int(self.scenario.get("runnumber")))
        
        self.allnodes = []
        self.availableresources = pongsim_resources.AvailableResources()
        numberofnodes = int(sm.get("nodes", 100))
        
        for c in range(numberofnodes):
            x = self.world.xsize/2.0*0.7*math.cos(float(c)/float(numberofnodes)*2.0*math.pi) + self.world.xsize/2.0
            y = self.world.ysize/2.0*0.7*math.sin(float(c)/float(numberofnodes)*2.0*math.pi) + self.world.ysize/2.0
            newnode = pongsim.Node(self.world, color = "#F00", x=x, y=y)
            if sm.get("scheme") == "ri":
                newapp = pongsim_ri_app.RIApplication(newnode)
            elif sm.get("scheme") == "floodsearch":
                newapp = pongsim_floodsearch_app.FloodSearchApplication(newnode)
            elif sm.get("scheme") == "floodindex":
                newapp = pongsim_floodindex_app.FloodIndexApplication(newnode)
            else:
                self.log.simerror("Unknown search scheme: "+sm.get("scheme"))
                self.finish = True
            self.allnodes.append(newnode)
        
        self.generate_topology(self.allnodes)
        
        self.initialresources = pongsim_resources.InitialResources(numberofnodes)
        mincopies = sm.get("mincopies", sm.get("copies", 1))
        maxcopies = sm.get("maxcopies", sm.get("copies", 1))
        if sm.get("indexaddinterval", None):
            numberofresources = int(numberofnodes * (sm.get("indexremoveinterval", sm.get("indexaddinterval")) / sm.get("indexaddinterval")) / ((mincopies+mincopies)/2.0))
        else:
            numberofresources = int(self.scenariomanager.get("nrofresources", 100))
        self.initialresources.generate(numberofresources, mincopies, maxcopies)
        for node in self.allnodes:
            node.application.initialresources(self.initialresources.get())
        
        for node in self.allnodes:
            node.counter.deactivate()
        self.scheduler.after(sm.get("startcollection", 0), self.startdatacollection, None)

    def generate_topology(self, nodes):
        """Generate the topology connecting nodes together according to the chosen scenario."""
        while 1:
            topology = pongsim_topologygenerator.Topology(len(nodes))
            topology.powerlaw_BA_degree(sm.get("avgdegree", 10))
            if topology.isconnected(): 
                break
            else:
                self.log.simstatus("Unconnected topology -> Regenerating")
        self.simcounter.add("real-avg-degree", topology.avgdegree())
        for nodeindex in range(len(nodes)):
            contacts = []
            for n in topology.neighbors(nodeindex):
                contacts.append(nodes[n])      # add first node of cluster as inbound node
            nodes[nodeindex].application.outofbanddata("contacts", contacts)

    def close(self):
        """Close the simulator and write the report."""
        if self.noreport: return
        reportname = self.outputdir + self.scenario.get("runoutput")
        fp = open(os.path.normcase(reportname), "w")

        sim_time = self.scheduler.time()

        fp.write(str(self.scenario))
        for node in self.allnodes:
            for counter in node.counter.counterlist():
                self.simcounter.add(counter, node.counter.sum(counter) / (sim_time-sm.get("startcollection", 0)))
            node.counter.close()
        self.simcounter.add("simulation-time", self.scheduler.time())
        fp.write(str(self.simcounter))
        self.simcounter.close()
        fp.close()
        
    def startdatacollection(self, dummy):
        """Start collecting the statistics."""
        for node in self.allnodes:
            node.counter.activate()

    def eachevent(self):
        """Check whether the simulation should end."""
        if self.scheduler.time() > sm.get("simtime", 5000): self.finish = True

####################################################################
###   M a i n   P r o g r a m
####################################################################

if __name__ == '__main__':
    pongsim.Simulator(GroupSimulation("./output/"), "./pongsim.log", "./scenarios/").run()
    
    

