"""
PONGsim (Python Overlay Network Graphical Simulator)
Module simulating standard flood searching

Version: 1.60

Nicklas Beijar, nbeijar@netlab.tkk.fi
Networking Laboratory, Helsinki University of Technology

######################################################################

Simulator for distribution algorithms for fully distributed user groups

Copyright (c) 2005-2007 Nicklas Beijar

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

######################################################################

To do:
- Nothing

"""

import random
import pongsim, pongsim_sil, pongsim_resources, pongsim_bloom

MSG_SEARCH_RESPONSE = 10
MSG_SEARCH_QUERY = 11

####################################################################
###   F l o o d I n d e x A p p l i c a t i o n
####################################################################

class FloodIndexApplication(pongsim_sil.SILApplication, pongsim_resources.Resources):

    def init(self):
        """Initialize the application."""
        pongsim_sil.SILApplication.init(self) # calling parent initiator is mandatory
        pongsim_resources.Resources.init(self, self.simulation.log, self.simulation.scenario.get("searchinterval", None), self.simulation.scenario.get("indexaddinterval", None), self.simulation.scenario.get("indexremoveinterval", self.simulation.scenario.get("indexaddinterval", None)), 
            self.simulation.scenario.get("removeinitialresources", True), self.simulation.scenario.get("searchtimeout", 20), self.simulation.scenario.get("dynamicresourcefrom", 0), self.simulation.scenario.get("dynamicresourceto", 
            self.simulation.scenario.get("nrofresources")), self.simulation.scenario.get("postponeremoval"), self.simulation.simcounter)
        self.lastadded = None
        self.nextupdate = self.simulation.scheduler.time()
        self.contacts = []
        self.remoteresources = {}
        
    ############ USER INTERFACE

    def getdisplayinfo(self):
        """Return the text written beside the node."""
        return str(self.node.address)

    def getdetailedinfo(self):
        """Return the the node's description that is written in a separate window when the node is examined."""
        t = pongsim_sil.SILApplication.getdetailedinfo(self)
        t = t + "Local resources:\n"
        t2 = "   "
        for resource in self.localresources:
            t2 = t2 + str(resource) + "  "
            if len(t2) > 30:
                t = t + t2
                t2 = "\n   "
        t = t + t2+"\n"    
        t = t + "Remote resources:\n"
        t2 = "   "
        for resourcekey in self.remoteresources.keys():
            t2 = t2 + str(self.remoteresources[resourcekey]) + "@" + str(resourcekey)
            if len(t2) > 30:
                t = t + t2
                t2 = "\n   "
        t = t + t2+"\n"    
        return t

    ############ GENERAL SUPPORT FUNCTIONS

    def outofbanddata(self, key, value):
        """Set the data that is agreed without any protocol (operator agreements) or manually configured.
        
        Corresponds to the user/administrator configuring the application. 
        """
        if key == "contacts":
            self.contacts = value
            for contact in self.contacts:
                self.addlink(contact, True, False, 256, None)
        else:
            self.simulation.log.simerror("Unknown key for outofbanddata: "+str(key))

    def receivegeneral(self, ttl, type, id, msg):
        """Handle reception of a general message and call the corresponding handler. 
        
        Called by the SIL framework.
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        if type == MSG_SEARCH_RESPONSE:
            self.on_receive_response(msg)
        elif type == MSG_SEARCH_QUERY:
            self.on_receive_query(msg)
        else:
            self.simulation.log.simerror("Unknown general message: "+str(type))

    ############ INDEX PROTOCOL

    def on_added_resource(self, resource):
        """Handle an added resource.
        
        Called by the Resources class when a resource has been added. 
        The floodindex application floods an update.
        """
        if self.simulation.scheduler.time() != self.lastadded:
            self.nextupdate = random.uniform(0.0, self.simulation.scenario.get("updatedelay")) + self.simulation.scheduler.time()
            self.simulation.scheduler.after(self.nextupdate - self.simulation.scheduler.time(), self.flood_index_update, resource)
            self.lastadded = self.simulation.scheduler.time()

    def on_removed_resource(self, resource):
        """Handle an removed resource.
        
        Called by the Resources class when a resource has been removed. 
        The floodindex application floods an update.
        """
        if self.simulation.scenario.get("skipupdateonremoval"): return
        if self.simulation.scheduler.time() != self.lastadded:
            self.nextupdate = random.uniform(0.0, self.simulation.scenario.get("updatedelay")) + self.simulation.scheduler.time()
            self.simulation.scheduler.after(self.nextupdate - self.simulation.scheduler.time(), self.flood_index_update, resource)
            self.lastadded = self.simulation.scheduler.time()

    def flood_index_update(self, resource=None):
        """Flood an index update."""
        if not self.node.world: return # scheduled functions need to check the node's existence
        if self.simulation.scheduler.time() < self.nextupdate: return # has been updated recently by other timer
        msg = (self.node, pongsim_bloom.BloomFilter(self.localresources, self.simulation.scenario.get("falsepositiveprob", 0.00)))
        self.floodindex(msg, ttl = int(self.simulation.scenario.get("indexttl", 256)))
        self.nextupdate = self.simulation.scheduler.time() + self.simulation.scenario.get("updateinterval")
        self.simulation.scheduler.after(self.nextupdate - self.simulation.scheduler.time(), self.flood_index_update, resource)

    def receiveindex(self, ttl, msg, destinations, firsthop = False):
        """Handle received index updates.
        
        Called by the SIL framework. 
        Determines the nodes to which the message is forwarded from the candidate destinations based on the index strategy. 
        """
        (peernode, peerbloom) = msg
        self.remoteresources[peernode] = peerbloom
        return [(msg, destinations), ]

    ############ SEARCH PROTOCOL

    def on_search_request(self, resource, searchid):
        """Handle a search request generated by the node. 
        
        Called by the Resources class when a resource request is generated.  
        """
        howmany = 0
        for peernode in self.remoteresources.keys():
            if self.remoteresources[peernode].check(resource):
                msg = (self.node, searchid, resource)
                self.sendgeneral(msg, peernode, MSG_SEARCH_QUERY)
                self.node.count("sent-searchquery")
                howmany += 1
        if howmany == 0:
            self.node.count("search-immediate-mismatch")

    def on_receive_query(self, msg):
        """Handle received search requests."""
        self.node.count("received-searchquery")
        (searcher, searchid, resource) = msg
        if resource in self.localresources:
            self.node.count("search-local-match")
            msg = (self.node, searchid, resource, True)
        else:
            self.node.count("search-local-mismatch")
            msg = (self.node, searchid, resource, False)
        self.sendgeneral(msg, searcher, MSG_SEARCH_RESPONSE)
        self.node.count("sent-searchrep")

    def on_receive_response(self, msg):
        """Handle a received search reply. 
        
        Count the answers. 
        """
        (where, searchid, resource, found) = msg
        self.node.count("received-searchrep")
        if found: 
            self.node.count("search-single-match")
            self.reportfound(searchid, where)
        else:
            self.node.count("search-single-mismatch")
            
            
if __name__ == '__main__':
    print "This is a framework for simulations, and is not runnable as such."
    print "To use it, create subclasses of the framework classes."
