"""
PONGsim (Python Overlay Network Graphical Simulator)
Module simulating standard flood searching

Version: 1.60

Nicklas Beijar, nbeijar@netlab.tkk.fi
Networking Laboratory, Helsinki University of Technology

######################################################################

Simulator for distribution algorithms for fully distributed user groups

Copyright (c) 2005-2007 Nicklas Beijar

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

######################################################################

To do:
- Nothing

"""

import pongsim, pongsim_sil, pongsim_resources

MSG_SEARCH_RESPONSE = 10

####################################################################
###   F l o o d S e a r c h A p p l i c a t i o n
####################################################################

class FloodSearchApplication(pongsim_sil.SILApplication, pongsim_resources.Resources):

    def init(self):
        """Initialize the application."""
        pongsim_sil.SILApplication.init(self) # calling parent initiator is mandatory
        pongsim_resources.Resources.init(self, self.simulation.log, self.simulation.scenario.get("searchinterval", None), self.simulation.scenario.get("indexaddinterval", None), self.simulation.scenario.get("indexremoveinterval", self.simulation.scenario.get("indexaddinterval", None)), 
            self.simulation.scenario.get("removeinitialresources", True), self.simulation.scenario.get("searchtimeout", 20), self.simulation.scenario.get("dynamicresourcefrom", 0), self.simulation.scenario.get("dynamicresourceto", 
            self.simulation.scenario.get("nrofresources")), self.simulation.scenario.get("postponeremoval"), self.simulation.simcounter)
        self.contacts = []
        
    ############ USER INTERFACE

    def getdisplayinfo(self):
        """Return the text written beside the node."""
        return str(self.node.address)

    def getdetailedinfo(self):
        """Return the the node's description that is written in a separate window when the node is examined."""
        t = pongsim_sil.SILApplication.getdetailedinfo(self)
        t = t + "Local resources:\n"
        t2 = "   "
        for resource in self.localresources:
            t2 = t2 + str(resource) + "  "
            if len(t2) > 30:
                t = t + t2
                t2 = "\n   "
        t = t + t2+"\n"    
        return t

    ############ GENERAL SUPPORT FUNCTIONS

    def outofbanddata(self, key, value):
        """Set the data that is agreed without any protocol (operator agreements) or manually configured.
        
        Corresponds to the user/administrator configuring the application. 
        """
        if key == "contacts":
            self.contacts = value
            for contact in self.contacts:
                self.addlink(contact, False, True, None, 256)
        else:
            self.simulation.log.simerror("Unknown key for outofbanddata: "+str(key))

    def receivegeneral(self, ttl, type, id, msg):
        """Handle reception of a general message and call the corresponding handler. 
        
        Called by the SIL framework.
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        if type == MSG_SEARCH_RESPONSE:
            self.on_receive_response(msg)
        else:
            self.simulation.log.simerror("Unknown general message: "+str(type))

    ############ SEARCH PROTOCOL

    def on_search_request(self, resource, searchid):
        """Handle a search request generated by the node. 
        
        Called by the Resources class when a resource request is generated.  
        """
        msg = (self.node, searchid, resource)
        self.floodsearch(msg, ttl = int(self.simulation.scenario.get("searchttl", 256)))

    def receivesearch(self, ttl, msg, destinations, firsthop = False):
        """Handle received search requests.
        
        Called by the SIL framework. 
        Determines the nodes to which the message is forwarded from the candidate destinations based on the search strategy. 
        """
        (searcher, searchid, resource) = msg
        if resource in self.localresources:
            self.replysearch(searcher, searchid, resource)
        return [(msg, destinations), ]

    def replysearch(self, searcher, searchid, resource):
        """Send a search reply."""
        msg = (self.node, searchid, resource)
        self.sendgeneral(msg, searcher, MSG_SEARCH_RESPONSE)
        self.node.count("sent-searchrep")

    def on_receive_response(self, msg):
        """Handle a received search reply."""
        (where, searchid, resource) = msg
        self.node.count("received-searchrep")
        self.reportfound(searchid, where)
            
if __name__ == '__main__':
    print "This is a framework for simulations, and is not runnable as such."
    print "To use it, create subclasses of the framework classes."
