"""
PONGsim (Python Overlay Network Graphical Simulator)
Resource module of the PONGsim framework

Version: 1.20

Nicklas Beijar, nbeijar@netlab.tkk.fi
Networking Laboratory, Helsinki University of Technology

######################################################################

PONGsim (Python Overlay Network Graphical Simulator)

Copyright (c) 2005-2007 Nicklas Beijar

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

######################################################################

To do:
- Everything

"""

import random

####################################################################
###   R e s o u r c e s
####################################################################

class Resources:

    """Node-specific class for generating resources and search requests.
    
    Should be subclassed together with an Application or SILApplication class. 
    """
    
    def init(self, log, searchinterval, indexaddinterval, indexremoveinterval, removeinitialresources, searchtimeout, dynamicresourcefrom, dynamicresourceto, postponeremoval, valuecounter):
        """Initialize the Resources module.
        
        Arguments:
        log - A Log for storing and displaying warning messages. 
        searchinterval - The average interval between successive search requests. 
        indexaddinterval -  The average interval between sucessive resource additions. 
        indexremoveinterval - The average time the resource is available at the node. 
        removeinitialresources - True if the initial resources should be removed after a indexremoveinterval.
        searchtimeout - The time period after which the search results are examined. 
        dynamicresourcefrom - The start of the number range of resource numbers added dynamically. 
        dynamicresourceto - The end of the number range of resources added dynamically. 
        postponeremoval - True if removal of resources should be delayed until the following addition of resources.
        valuecounter - A StatisticsCounter that counts the delays and number of replies. 
        """
        self.localresources = []
        self.toremove = []
        self.pendingsearches = {}
        self.pendingsearchtimes = {}
        self.pendingsearchsendtime = {}
        self.searchinterval = searchinterval
        self.indexaddinterval = indexaddinterval
        self.indexremoveinterval = indexremoveinterval
        self.removeinitialresources = removeinitialresources
        self.searchtimeout = searchtimeout
        self.dynamicresourcefrom = dynamicresourcefrom
        self.dynamicresourceto = dynamicresourceto
        self.postponeremoval = postponeremoval
        self.valuecounter = valuecounter
        self.log = log
        
        self.simulation.scheduler.after(random.expovariate(1.0/self.searchinterval), self.generate_search, None) 
        if self.indexaddinterval:
            self.simulation.scheduler.after(random.expovariate(1.0/self.indexaddinterval), self.addresource, None)
        
    ############ RESOURCES

    def initialresources(self, resources):
        """Add the given resources as the initial resources."""
        for resource in resources:
            self.localresources.append(resource)
            self.node.world.simulation.availableresources.register(resource, self.node)
            if self.removeinitialresources: self.simulation.scheduler.after(random.expovariate(1.0/self.indexremoveinterval), self.removeresource, resource)
            self.on_added_resource(resource)

    def addresource(self, dummy=None):
        """Add a resource to the node. 
        
        Adds a random resource, id defined by the scenario-parameters dynamicresourcefrom and dynamicresourceto.
        If dynamicresourceto is undefined, nrofresources is used as upper limit. 
        Schedules itself again after interval defined in the scenario-parameter indexaddinterval.
        Schedules removal of the added resource after interval defined in the scenario-parameter indexremoveinterval
        (or in indexaddinterval if indexremoveinterval is undefined).
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        newresource = random.randint(self.dynamicresourcefrom, self.dynamicresourceto)
        self.localresources.append(newresource)
        self.node.world.simulation.availableresources.register(newresource, self.node)
        self.node.count("dyn-resource-add")
        self.simulation.scheduler.after(random.expovariate(1.0/self.indexremoveinterval), self.removeresource, newresource)
        self.simulation.scheduler.after(random.expovariate(1.0/self.indexaddinterval), self.addresource, None)
        self.on_added_resource(newresource)
        for resource in self.toremove:
            self.localresources.remove(resource)
            self.node.world.simulation.availableresources.deregister(resource, self.node)
            self.node.count("dyn-resource-remove")
            self.on_removed_resource(resource)
        self.toremove = []


    def removeresource(self, resource):
        """Remove a resource from the node. """
        if not self.node.world: return # scheduled functions need to check the node's existence
        if resource not in self.localresources: 
            self.log.simwarning("Removal of nonexistent resource")
            return
        if self.postponeremoval:
            self.toremove.append(resource)
        else:
            self.localresources.remove(resource)
            self.node.world.simulation.availableresources.deregister(resource, self.node)
            self.node.count("dyn-resource-remove")
            self.on_removed_resource(resource)

    def on_added_resource(self, resource):
        """Handles an added resource.
        
        Called by the Resources class when a resource has been added. The Application can then
        perform necessary functions, such as distributing index updates. To be subclassed. 
        """
        pass

    def on_removed_resource(self, resource):
        """Handles an removed resource.
        
        Called by the Resources class when a resource has been removed. The Application can then
        perform necessary functions, such as distributing index updates. To be subclassed. 
        """
        pass

    ############ SEARCHING

    nextsearchid = 1

    def generate_search(self, dummy=None):
        """Generate a search request, which is flooded on the search links. 
        
        Scheduled calling, with rescheduling. 
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        resource = self.node.world.simulation.availableresources.random(self.node)
        self.node.count("generated-searches")
        searchid = Resources.nextsearchid
        Resources.nextsearchid += 1
        self.pendingsearches[searchid] = []
        self.pendingsearchtimes[searchid] = None
        self.pendingsearchsendtime[searchid] = self.simulation.scheduler.time()
        self.simulation.scheduler.after(self.searchtimeout, self.timeoutsearch, searchid) # schedule timeout for checking results
        self.simulation.scheduler.after(random.expovariate(1.0/self.searchinterval), self.generate_search, None)  # schedule next search
        self.on_search_request(resource, searchid)
        
    def on_search_request(self, resource, searchid):
        """Handles a search request generated by the node. 
        
        Called by the Resources class when a resource request is generated. The application can then e.g. 
        forward the request to other nodes. To be subclassed. 
        """
        pass

    def reportfound(self, searchid, where):
        if self.pendingsearches.has_key(searchid):
            if where not in self.pendingsearches[searchid]:
                self.node.count("search-single-match-unique")
                self.pendingsearchtimes[searchid] = self.simulation.scheduler.time()-self.pendingsearchsendtime[searchid]
            self.pendingsearches[searchid].append(where)
            self.valuecounter.add("search-delay", self.simulation.scheduler.time()-self.pendingsearchsendtime[searchid])
            if len(self.pendingsearches[searchid]) == 1:
                self.valuecounter.add("search-delay-first", self.simulation.scheduler.time()-self.pendingsearchsendtime[searchid])
        else:
            self.log.simwarning("Result arrived for nonexistent or timed out search. It's OK, but statistics not collected. Try increasing the timeout.")

    def timeoutsearch(self, searchid):
        """Examine the search results collected during the search, and collect statistics about the search.
        
        Called when a given time has expired since the search was initiated. A search request timed out with either results or without results. 
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        if (not self.pendingsearches.has_key(searchid)) or (not self.pendingsearchtimes.has_key(searchid)):
            self.log.simwarning("Search timeout for nonexistent search")
            return
        if len(self.pendingsearches[searchid]) > 0:
            self.node.count("search-result-found")
            self.valuecounter.add("search-responses", len(self.pendingsearches[searchid]))
            self.valuecounter.add("search-delay-last", self.pendingsearchtimes[searchid])
            unique = []
            for node in self.pendingsearches[searchid]:
                if node not in unique: unique.append(node)
            self.valuecounter.add("search-found-copies", len(unique))
        else:
            self.node.count("search-result-not-found")
        del self.pendingsearches[searchid]
        del self.pendingsearchtimes[searchid]
        del self.pendingsearchsendtime[searchid]

####################################################################
#####   I n i t i a l R e s o u r c e s
####################################################################

class InitialResources:
    
    """A class for generating the resources that initially are available."""

    def __init__(self, nodes):
        """Initialize the InitialResources for the given number of nodes."""
        self.nodes = nodes
        self.noderesources = []
        for i in range(nodes):
            self.noderesources.append([])
        
    def get(self):
        """Return the initial resources for the node that calls this function."""
        return self.noderesources.pop(0)
        
    def generate(self, numberofresources, mincopies=1, maxcopies=1):
        """Generate the given number of resources and allocate them to the nodes.
        
        The number of copies of a resource is random between the given limits.
        """
        for resource in range(numberofresources):
            copies = random.randint(mincopies, maxcopies)
            given = []
            for i in range(copies):
                while 1:
                    node = random.randint(0, self.nodes-1)
                    if node not in given: break
                given.append(node)
                self.noderesources[node].append(resource)

####################################################################
#####   A v a i l a b l e R e s o u r c e s
####################################################################

class AvailableResources:
    
    """A simulator-wide register for the resources available at different nodes.
    
    Used to assist searching so that a search will be directed to a resource that
    1) exists 
    2) is at any node.
    """
    
    def __init__(self):
        """Init the AvailableResrouce storage."""
        self.resources = {}
        
    def register(self, resource, node):
        """Register a given resource that is available at the given node."""
        if not self.resources.has_key(resource):
            self.resources[resource] = []
        self.resources[resource].append(node)

    def deregister(self, resource, node):
        """Deregister a given resource that was available at the given node."""
        if not self.resources.has_key(resource): 
            self.log.simwarning("Deregistering nonexistent resource")
            return
        if node not in self.resources[resource]: 
            self.log.simwarning("Deregistering resource not in correct node")
            return
        self.resources[resource].remove(node)
        if len(self.resources[resource]) == 0: 
            del self.resources[resource] # this bug was hard to find!
        
    def random(self, exclude=None):
        """Returns a random resource that can be available at any node.
        
        If argument exclude is given, a resource found in the excluded node will not be selected.
        """
        resources = self.resources.keys()
        while 1:
            chosen = resources[random.randint(0, len(resources)-1)]
            if (not exclude) or (exclude not in self.resources[chosen]): break
        return chosen

if __name__ == '__main__':
    print "This is a framework for simulations, and is not runnable as such."
    print "To use it, create subclasses of the framework classes."
