"""
PONGsim (Python Overlay Network Graphical Simulator)
Module simulating the Routed Index approach

Version: 1.60

Nicklas Beijar, nbeijar@netlab.tkk.fi
Networking Laboratory, Helsinki University of Technology

######################################################################

Simulator for distribution algorithms for fully distributed user groups

Copyright (c) 2005-2007 Nicklas Beijar

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

######################################################################

To do:
- Everything

"""

import random, bisect
import pongsim, pongsim_sil, pongsim_resources, pongsim_bloom

MSG_UPDATE = 10
MSG_ACK = 11
MSG_SEARCH_QUERY = 12
MSG_SEARCH_RESPONSE = 13

####################################################################
###   O v e r l a y T o p o l o g y
####################################################################

class OverlayTopology:
    
    """A class for storing one node's view of the group's topology.
    
    Stores the link table of the group, which is used to calculate a routing table. 
    """
    
    def __init__(self, node):
        """Initialize the topology rooted in the given node."""
        self.myself = node
        self.links = {}
        self.updated = True
        self.distance = {}
        self.distance[self.myself] = 0
        
    
    def update(self, node, neighbors):
        """Update the link table with the given node's given neighbors."""
        if (self.links.has_key(node)) and (len(self.links[node]) == len(neighbors)):
            self.updated = False
            for neighbor in self.links[node]:
                if neighbor not in neighbors: self.updated = True
        else:
            self.updated = True
        self.links[node] = neighbors
        
    def isreachable(self, node):
        """Return True if the given node is reachable, otherwise False."""
        if self.updated: self.calculate()
        return self.distance.has_key(node)

    def distance(self, node):
        """Return the distance to the given node."""
        if self.updated: self.calculate()
        if self.distance.has_key(node):
            return self.distance[node]
        return None

    def calculate(self):
        """Calculate a new routing table using Dijkstra's algorithm."""
        self.distance = {}
        self.distance[self.myself] = 0
        E = [self.myself]
        O = []
        for node in self.links[self.myself]:
            O.append((node, 1))
        while 1:
            if len(O) == 0: break
            (V, D) = O.pop(0)
            if D < 0: break
            if V in E: continue
            E.append(V)
            self.distance[V] = D
            if self.links.has_key(V):
                for dest in self.links[V]:
                    bisect.insort(O, (dest, D+1))
        self.updated = False

    def __str__(self):
        """Return a multi-line string representation of the routing table."""
        if self.updated: self.calculate()
        s = ""
        for node in self.distance.keys():
            s += str(node.getaddress()) + ": "
            s += str(self.distance[node]) + "\n"
        return s

####################################################################
###   R I A p p l i c a t i o n
####################################################################

class RIApplication(pongsim_sil.SILApplication, pongsim_resources.Resources):

    def init(self):
        """Initialize the application."""
        pongsim_sil.SILApplication.init(self) # calling parent initiator is mandatory
        pongsim_resources.Resources.init(self, self.simulation.log, self.simulation.scenario.get("searchinterval", None), self.simulation.scenario.get("indexaddinterval", None), self.simulation.scenario.get("indexremoveinterval", self.simulation.scenario.get("indexaddinterval", None)), 
            self.simulation.scenario.get("removeinitialresources", True), self.simulation.scenario.get("searchtimeout", 20), self.simulation.scenario.get("dynamicresourcefrom", 0), self.simulation.scenario.get("dynamicresourceto", 
            self.simulation.scenario.get("nrofresources")), self.simulation.scenario.get("postponeremoval"), self.simulation.simcounter)
        self.contacts = []
        self.nextupdate = {}
        self.remoteresources = {}
        self.lastadded = None
        self.topology = OverlayTopology(self.node)
        
    ############ USER INTERFACE

    def getdisplayinfo(self):
        """Return the text written beside the node."""
        return str(self.node.address)

    def getdetailedinfo(self):
        """Return the the node's description that is written in a separate window when the node is examined."""
        t = pongsim_sil.SILApplication.getdetailedinfo(self)
        t = t + "Local resources:\n"
        t2 = "   "
        for resource in self.localresources:
            t2 = t2 + str(resource) + "  "
            if len(t2) > 30:
                t = t + t2
                t2 = "\n   "
        t = t + t2+"\n"         
        t = t + "Remote resources:\n"
        t2 = "   "
        for resourcekey in self.remoteresources.keys():
            t2 = t2 + str(self.remoteresources[resourcekey]) + "@" + str(resourcekey)
            if len(t2) > 30:
                t = t + t2
                t2 = "\n   "
        t = t + t2+"\n"    
        t += "Routing table:\n"
        t += str(self.topology).replace("\n", "   ")
        return t

    ############ GENERAL SUPPORT FUNCTIONS

    def outofbanddata(self, key, value):
        """Set the data that is agreed without any protocol (operator agreements) or manually configured.
        
        Corresponds to the user/administrator configuring the application. 
        """
        if key == "contacts":
            self.contacts = value
            for contact in self.contacts:
                self.nextupdate[contact] = random.uniform(0.0, self.simulation.scenario.get("updatedelay")) + self.simulation.scheduler.time()
                self.simulation.scheduler.after(self.nextupdate[contact] - self.simulation.scheduler.time(), self.generate_update, contact)
            self.topology.update(self.node, self.contacts)
        else:
            self.simulation.log.simerror("Unknown key for outofbanddata: "+str(key))

    def receivegeneral(self, ttl, type, id, msg):
        """Handle reception of a general message and call the corresponding handler. 
        
        Called by the SIL framework.
        """
        if not self.node.world: return # scheduled functions need to check the node's existence
        if type == MSG_UPDATE:
            self.on_receive_update(msg)
        elif type == MSG_ACK:
            self.on_receive_ack(msg)
        elif type == MSG_SEARCH_QUERY:
            self.on_receive_query(msg)
        elif type == MSG_SEARCH_RESPONSE:
            self.on_receive_response(msg)
        else:
            self.simulation.log.simerror("Unknown general message: "+str(type))

    ############ UPDATE EXCHANGE PROTOCOL
    
    UPDATE_REQUEST = 1
    UPDATE_RESPONSE = 2
    
    def generate_update(self, node):
        """Generate and send an update message to the given node."""
        if not self.node.world: return # scheduled functions need to check the node's existence
        if self.simulation.scheduler.time() < self.nextupdate[node]: return # has been updated recently by other timer
        if not self.topology.isreachable(node):
            self.node.count("x-unreachable-on-send")
            del self.nextupdate[node]
            return
        # single group so far
        msg = (self.node, self.contacts, pongsim_bloom.BloomFilter(self.localresources, self.simulation.scenario.get("falsepositiveprob", 0.00)), self.UPDATE_REQUEST)
        self.node.count("sent-update")
        self.sendgeneral(msg, node, MSG_UPDATE)
        self.nextupdate[node] = self.simulation.scheduler.time() + self.simulation.scenario.get("updateinterval")
        self.simulation.scheduler.after(self.nextupdate[node] - self.simulation.scheduler.time(), self.generate_update, node)

    def on_receive_update(self, msg):
        """Handle a received update message."""
        self.node.count("received-update")
        (peernode, peercontacts, peerbloom, updatetype) = msg

        # Don't accept updates from unknown nodes.
        if not self.topology.isreachable(peernode):
            self.node.count("x-unreachable-on-receive")
            return

        # Update info about sender
        self.remoteresources[peernode] = peerbloom
        self.nextupdate[peernode] = self.simulation.scheduler.time() + self.simulation.scenario.get("updateinterval")
        self.topology.update(peernode, peercontacts)

        # Learn new remote contacts
        for contact in peercontacts:
            if not self.nextupdate.has_key(contact):
                self.nextupdate[contact] = random.uniform(0.0, self.simulation.scenario.get("updatedelay")) + self.simulation.scheduler.time()
                self.node.count("new-indirect-contacts")
                self.simulation.scheduler.after(self.nextupdate[contact] - self.simulation.scheduler.time(), self.generate_update, contact)
                
        # Send reply
        # single group so far
        if updatetype == self.UPDATE_REQUEST:
            msg = (self.node, self.contacts, pongsim_bloom.BloomFilter(self.localresources, self.simulation.scenario.get("falsepositiveprob", 0.00)), self.UPDATE_RESPONSE)
            self.node.count("sent-update")
            self.sendgeneral(msg, peernode, MSG_UPDATE)
        else:
            msg = (self.node)
            self.node.count("sent-ack")
            self.sendgeneral(msg, peernode, MSG_ACK)
            
        self.simulation.scheduler.after(self.nextupdate[peernode] - self.simulation.scheduler.time(), self.generate_update, peernode)
        
    def on_receive_ack(self, msg):
        """Handle a received ack message."""
        self.node.count("received-ack")
        
    ############ DYNAMIC RESOURCES

    def on_added_resource(self, resource):
        """Handle an added resource.
        
        Called by the Resources class when a resource has been added. 
        The RI application forces an update to all known peers.
        """
        if self.simulation.scheduler.time() != self.lastadded:
            self.forceupdate()
            self.lastadded = self.simulation.scheduler.time()

    def on_removed_resource(self, resource):
        """Handle an removed resource.
        
        Called by the Resources class when a resource has been removed. 
        The RI application forces an update to all known peers.
        """
        if self.simulation.scenario.get("skipupdateonremoval"): return
        if self.simulation.scheduler.time() != self.lastadded:
            self.forceupdate()
            self.lastadded = self.simulation.scheduler.time()
    
    def forceupdate(self):
        """Force an update to all known peers as a result of index change.
        
        No need to schedule updates to peers that anyway will receive an update soon.
        """
        for peernode in self.nextupdate.keys():
            if (self.nextupdate[peernode] - self.simulation.scheduler.time()) > self.simulation.scenario.get("updatedelay"):
                self.nextupdate[peernode] = random.uniform(0.0, self.simulation.scenario.get("updatedelay")) + self.simulation.scheduler.time()
                self.simulation.scheduler.after(self.nextupdate[peernode] - self.simulation.scheduler.time(), self.generate_update, peernode)

    ############ SEARCH PROTOCOL

    def on_search_request(self, resource, searchid):
        """Handles a search request generated by the node. 
        
        Called by the Resources class when a resource request is generated.  
        """
        howmany = 0
        for peernode in self.remoteresources.keys():
            if self.remoteresources[peernode].check(resource):
                msg = (self.node, searchid, resource)
                self.sendgeneral(msg, peernode, MSG_SEARCH_QUERY)
                self.node.count("sent-searchquery")
                howmany += 1
        if howmany == 0:
            self.node.count("search-immediate-mismatch")

    def on_receive_query(self, msg):
        """Handle received search requests."""
        self.node.count("received-searchquery")
        (searcher, searchid, resource) = msg
        if resource in self.localresources:
            self.node.count("search-local-match")
            msg = (self.node, searchid, resource, True)
        else:
            self.node.count("search-local-mismatch")
            msg = (self.node, searchid, resource, False)
        self.sendgeneral(msg, searcher, MSG_SEARCH_RESPONSE)
        self.node.count("sent-searchrep")

    def on_receive_response(self, msg):
        """Handle a received search reply. 
        
        Count the answers. 
        """
        (where, searchid, resource, found) = msg
        self.node.count("received-searchrep")
        if found: 
            self.node.count("search-single-match")
            self.reportfound(searchid, where)
        else:
            self.node.count("search-single-mismatch")

if __name__ == '__main__':
    print "This is a framework for simulations, and is not runnable as such."
    print "To use it, create subclasses of the framework classes."
