"""
PONGsim (Python Overlay Network Graphical Simulator)
Module implementing the SIL-model

Version: 7.20

Nicklas Beijar, nbeijar@netlab.tkk.fi
Networking Laboratory, Helsinki University of Technology

######################################################################

PONGsim (Python Overlay Network Graphical Simulator)

Copyright (c) 2005-2007 Nicklas Beijar

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use,
copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following
conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

######################################################################

To do:
- Improve comments

"""

import pongsim, pongsim_sil
import random

try:
    from Tkinter import *
except ImportError:
    pass


SIL_LINK_SEARCH = 0
SIL_LINK_INDEX = 1
SIL_LINK_SEARCH_TTL = 2
SIL_LINK_INDEX_TTL = 3
SIL_LINK_NODE = 4

SIL_MSG_SEARCH = 2
SIL_MSG_INDEX = 1
    
######################################################################
#####    S I L A P P L I C A T I O N
######################################################################

class SILApplication(pongsim.Application):

    """Framework for applications based on the Search/Index Link model."""

    def init(self):
        """Initiate the SILApplication.
        
        May be subclassed. Subclass MUST call the parent class. 
        """
        self.seencache = []
        self.seencachesize = 400
        self.ctr_searchforwards = pongsim.Counter("searchforwards")
        self.links = []
 
    ##### HANDLERS FOR MESSAGES IMPLEMENTED BY SUBCLASS

    def receiveindex(self, ttl, msg, destinations):
        """Handle received index requests.
        
        Determines the nodes to which the message is forwarded from the candidate destinations based on 
        the index strategy. Returns a list of (msg, destination) tuples, where msg is the (possibly modified) 
        message to be forwarded to the nodes in the destination list. To be subclassed. 
        """
        return [(msg, destinations), ]
        
    def receivesearch(self, ttl, msg, destinations):
        """Handle received search requests.
        
        Determines the nodes to which the message is forwarded from the candidate destinations based on 
        the search strategy. Returns a list of (msg, destination) tuples, where msg is the (possibly modified) 
        message to be forwarded to the nodes in the destination list. To be subclassed. 
        """
        return [(msg, destinations), ]

    def receivegeneral(self, ttl, type, id, msg):
        """Handle received message of other type.
        
        User-defined messages are classed according to their type. To be subclassed.
        """
        return (msg, destinations)

    ##### UTILITY

    def isnew(self, id):
        """Returns True if the message with the given id has been forwarded earlier by this node.
        
        Remembers the current id. NOT recommended to be subclassed, but CAN be subclassed to skip checking.
        """
        if id in self.seencache: return False
        self.seencache.append(id)
        if len(self.seencache) > self.seencachesize: del self.seencache[0]
        return True
        
    ##### SENDING, RECEIVING, FORWARDING MESSAGES
        
    def receive(self, packedmsg):
        """Handle received message.
        
        This function is called by the Node when a message is received. The SIL framework handles
        index, search and generic messages separately. NOT to be subclassed. 
        """
        (ttl, type, id, msg, prevnode) = packedmsg
        if type == SIL_MSG_INDEX: # index
            self.node.count("received-index-msg")
            if self.isnew(id):
                self.node.count("received-unique-index-msg")
                destinations = []
                for link in self.links:
                    if link[SIL_LINK_INDEX]:
                        if link[SIL_LINK_NODE] != prevnode and link[SIL_LINK_NODE] != self.node:
                            destinations.append(link[SIL_LINK_NODE])
                forwards = self.receiveindex(ttl, msg, destinations)
                for forward in forwards:
                    (msg, destinations) = forward
                    if destinations and ttl > 1:
                        self.node.count("forwarded-index-msg", len(destinations))
                        self.forwardindex(msg, destinations, ttl-1, id)
        elif type == SIL_MSG_SEARCH: # search
            self.node.count("received-search-msg")
            if self.isnew(id):
                self.node.count("received-unique-search-msg")
                destinations = []
                for link in self.links:
                    if link[SIL_LINK_SEARCH]:
                        if link[SIL_LINK_NODE] != prevnode and link[SIL_LINK_NODE] != self.node: 
                            destinations.append(link[SIL_LINK_NODE])
                firsthop = prevnode == self.node
                forwards = self.receivesearch(ttl, msg, destinations, firsthop)
                for forward in forwards:
                    (msg, destinations) = forward
                    if destinations and ttl > 1:
                        self.node.count("forwarded-search-msg", len(destinations))
                        self.forwardsearch(msg, destinations, ttl-1, id)
        else: # general
            self.node.count("received-general-msg")
            self.receivegeneral(ttl, type, id, msg)

    def floodsearch(self, msg, ttl = None, id = None):
        """Send a search message using flooding. 
        
        If no ttl is specified, the link-specific one is used. If no id is specified, a random one is generated. 
        NOT to be subclassed.
        """
        if id == None: id = random.randint(0, 2147483645) # generate a random id if none is given
        self.node.count("sent-search-msg")
        self.receive((ttl+1, SIL_MSG_SEARCH, id, msg, self.node))
    
    def forwardsearch(self, msg, destinations, ttl = None, id = None):
        """Forward a previously received search message using flooding.
        
        If no ttl is specified, the link-specific one is used. NOT to be subclassed.
        """
        for dest in destinations:
            linkttl = ttl
            if linkttl == None: 
                for link in self.links:
                    if link[SIL_LINK_SEARCH]:
                        if link[SIL_LINK_NODE]==dest:
                            linkttl = link[SIL_LINK_INDEX_TTL]
            self.node.send((linkttl, SIL_MSG_SEARCH, id, msg, (self.node,)), dest)
    
    def floodindex(self, msg, ttl = None, id = None):
        """Send a index update message using flooding. 
        
        If no ttl is specified, the link-specific one is used. If no id is specified, a random one is generated. 
        NOT to be subclassed.
        """
        if id == None: id = random.randint(-30000, 30000) # generate a random id if none is given
        self.node.count("sent-index-msg")
        self.receive((ttl+1, SIL_MSG_INDEX, id, msg, self.node))

    def forwardindex(self, msg, destinations, ttl = None, id = None):
        """Forward a previously received index update message using flooding.
        
        If no ttl is specified, the link-specific one is used. NOT to be subclassed.
        """
        for dest in destinations:
            linkttl = ttl
            if linkttl == None: 
                for link in self.links:
                    if link[SIL_LINK_INDEX]:
                        if link[SIL_LINK_NODE]==dest:
                            linkttl = link[SIL_LINK_INDEX_TTL]
            self.node.send((linkttl, SIL_MSG_INDEX, id, msg, (self.node,)), dest)
    
    def sendgeneral(self, msg, dest, type = 0, ttl = None, id = None):
        """Send a message of other type to a specified node.
        
        Messages are sent directly to the destination - there is no need for a search/index link path to the destination. 
        TTL and ID defaults to None, since these messages are not affected by flooding. They can be used for 
        an application-specific purpose. Type can be any integer, except for the reserved values (1 and 2), defaulting to 0. 
        NOT recommended to be subclassed.
        """
        self.node.send((ttl, type, id, msg, (self.node,)), dest)

    ##### LINK MANAGEMENT

    def addlink(self, dest, index, search, indexttl = 255, searchttl = 255):
        """Add a search or index link (or both) from this node to the destination node.  

        For an index link: index = True, indexttl defined (default 255)
        For a search link: search = True, searchttl defined (default 255)
        If there already exists a link to the given node, that is updated. 
        """
        assert index or search
        isanewlink = True
        for i in range(len(self.links)):
            if self.links[i][3] == dest:
                isanewlink = False
                if searchttl == None: searchttl = self.links[i][2]
                if indexttl == None: indexttl = sefl.links[i][3]
                self.links[i] = (self.links[i][0] or search, self.links[i][1] or index, searchttl, indexttl, dest)
        if isanewlink: self.links.append((search, index, searchttl, indexttl, dest))
    
    def removelink(self, dest, index, search):
        """Removes a search or index link (or both) from the node to the destination node.
        
        Not implemented. 
        """
        assert False # not implemented
    
    ##### USER INTERFACE
    
    def draw(self, canvas, x, y):
        """Draw the application-specific details - here the search and index links."""
        for link in self.links:
            if link[0] and link[1]: # search and index
                fill = "#000"
            elif link[0]:               # search
                fill = "#F00"
            elif link[1]:               # index
                fill = "#00F"
            else:                       # no link
                fill = "#FFF"
            destx = link[4].getx()
            desty = link[4].gety()
            if not link[4].world:       # link to inexistent node
                destx = x + 40
                desty = y - 40
            canvas.create_line(x, y, link[4].getx(), link[4].gety(), fill=fill, width=1, arrow=LAST)

    def getdetailedinfo(self):
        """Return the the node's description that is written in a separate window when the node is examined."""
        t = pongsim.Application.getdetailedinfo(self)
        t = t + "Search links:\n"
        for link in self.links:
            if link[pongsim_sil.SIL_LINK_SEARCH]:
                t = t + "   " + str(link[pongsim_sil.SIL_LINK_NODE].address) + "\n"
        t = t + "Index links:\n"
        for link in self.links:
            if link[pongsim_sil.SIL_LINK_INDEX]:
                t = t + "   " + str(link[pongsim_sil.SIL_LINK_NODE].address) + "\n"
        return t
        
if __name__ == '__main__':
    print "This is a framework for p2p simulations, and is not runnable as such."
    print "To use it, create subclasses of the framework classes."
    